import os
import unittest

import fsl.HookHandler
import tutils

# See notes in test_interpreter.py on the following lines.
try:
    set, frozenset
except NameError:
    import sets
    set = sets.Set
    frozenset = sets.ImmutableSet

class TesterHookHandler(fsl.HookHandler.HookHandler, unittest.TestCase):
    def __init__(self):
        self.reset_hooks()

    def reset_hooks(self):
        self.include_fileset = set()     # set(filename)
        self.exclude_fileset = set()     # set(filename)
        self.scan_directory_dict = {} # {dirname: (number of starts, number of ends)
        self.source_msg = ''

    def include_file(self, fileobj, rule):
        self.failIf(not fileobj, '"%s": Include-file hook: fileobj is empty' % self.source_msg)
        self.failIf(not rule, '"%s": Include-file hook: rule is empty' % self.source_msg)
        filename = fileobj.abs_filename
        self.failIf(not os.path.isfile(filename),
            '"%s": Include-file hook: file "%s" is not a valid file' % (self.source_msg, filename))
        if filename in self.include_fileset:
            self.fail('"%s": Include-file hook called twice for file "%s"' % (self.source_msg, filename))
        if filename in self.exclude_fileset:
            self.fail('"%s": IFH: Both exclude-file and include-file hooks called for "%s"' % (self.source_msg, filename))
        self.include_fileset.add(filename)

    def exclude_file(self, fileobj, rule):
        self.failIf(not fileobj, '"%s": Exclude-file hook: filename is empty')
        self.failIf(not rule, '"%s": Exclude-file hook: rule is empty')
        filename = fileobj.abs_filename
        if filename in self.exclude_fileset:
            self.fail('"%s": Exclude-file hook called twice for file "%s"' % (self.source_msg, filename))
        if filename in self.include_fileset:
            self.fail('"%s": EFH: Both exclude-file and include-file hooks called for "%s"' % (self.source_msg, filename))
        self.exclude_fileset.add(filename)

    def start_scan(self, dirnameobj, glob_pattern):
        dirname = dirnameobj.abs_filename
        self.failIf(not dirname, '"%s": Start-scan hook: dirname is empty' % self.source_msg)
        self.failIf(not glob_pattern, '"%s": Start-scan hook: glob pattern is empty' % self.source_msg)
        self.failIf(not os.path.isdir(dirnameobj.abs_filename),
            '"%s": Start-scan hook: dirname "%s" is not a valid directory' % (self.source_msg, dirname))

        starts, ends = self.scan_directory_dict.get(dirname, (0, 0))
        self.scan_directory_dict[dirname] = (starts+1, ends)

    def end_scan(self, dirnameobj, files_count, bytes_count):
        dirname = dirnameobj.abs_filename
        self.failIf(not dirname, '"%s": End-scan hook: dirname is empty' % self.source_msg)
        self.failIf(not os.path.isdir(dirnameobj.abs_filename),
            '"%s": End-scan hook: dirname "%s" is not a valid directory' % (self.source_msg, dirname))

        self.failIf(files_count<0, '"%s": End-scan hook: files-count < 0' % self.source_msg)
        self.failIf(bytes_count<0, '"%s": End-scan hook: bytes-count < 0' % self.source_msg)
        self.failIf(files_count==0 and bytes_count>0, '"%s": End-scan hook: bytes > 0 but files = 0' % self.source_msg)

        if dirname not in self.scan_directory_dict:
            self.fail('"%s": End-scan hook called for %s, but start-scan hook was not called' % (self.source_msg, dirname) )

        starts, ends = self.scan_directory_dict[dirname]
        self.scan_directory_dict[dirname] = (starts, ends+1)
